// $Id: irc.cc,v 1.44 2003/01/15 23:43:46 toast Exp $

// C headers
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include "oomon.h"

// C++ headers
#include <string>
#include <vector>

#include "strtype"
#include "config.h"
#include "log.h"
#include "irc.h"
#include "main.h"
#include "util.h"
#include "engine.h"

bool   IRC::connected = false;
int    IRC::plug = 0;
bool   IRC::AmIAnOper = false;
std::string IRC::Server_Name = "";
bool   IRC::GettingCASVCSReports = false;
bool   IRC::GettingKlines = false;
bool   IRC::GettingDlines = false;
std::string IRC::MyNick = "";
KlineList IRC::Klines('K');
KlineList IRC::Dlines('D');
unsigned long IRC::HostAddr = 0;
unsigned long IRC::myaddr = 0;

int IRC::Connect()
{
  struct sockaddr_in	socketname;
  struct sockaddr_in	localaddr;
  struct hostent	*remote_host;
  struct hostent	*local_host;
  char			buf[MAX_BUFF];
  char			s[MAX_BUFF];
  int			portnum = 6667;
  char			server[MAX_HOST],
			*hold;
  socklen_t		size;
  unsigned long		remoteaddr;

  #ifdef IRC_DEBUG
    cout << "Connecting to IRC server..." << endl;
  #endif

  Log::Write("Connecting to IRC");

  if ((plug = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    cerr << "Error: Can't assign fd for socket!" << endl;
    return INVALID;
  }

  int optval = 1;
  setsockopt(plug, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval));
  setsockopt(plug, SOL_SOCKET, SO_KEEPALIVE, &optval, sizeof(optval));

  (void)bzero((char *)&socketname, sizeof(socketname));

  (void)bzero((char *)&localaddr, sizeof(struct sockaddr_in));
  if (Config::GetHostName() != "")
    if (local_host = gethostbyname(Config::GetHostName().c_str()))
      (void)bcopy((char *) local_host->h_addr,
	(char *) &localaddr.sin_addr,
	local_host->h_length);
  localaddr.sin_family = AF_INET;
  localaddr.sin_port = 0;

  if (bind(plug, (struct sockaddr *) &localaddr, sizeof(socketname)) < 0)
    cerr << "Error: Unable to bind virtual host" << endl;

  Server_Name = Config::GetServerHostName();

  if ((struct hostent *) 0 ==
    (remote_host = gethostbyname(Server_Name.c_str())))
  {
    cerr << "Error: Unknown host: " << Server_Name << endl;
    return INVALID;
  }
  (void)bcopy((char *) remote_host->h_addr,
	(char *) &socketname.sin_addr,
	remote_host->h_length);
  HostAddr = socketname.sin_addr.s_addr;
  socketname.sin_family = AF_INET;
  socketname.sin_port = htons(Config::GetServerPort());

  while (connect(plug, (struct sockaddr *) &socketname, sizeof(socketname)) < 0)
    if (errno != EINTR) {
      cerr << "Error: Connect " << errno << " (" << strerror(errno) << ")" << endl;
      return INVALID;
    }

  connected = true;

  size = sizeof(socketname);
  getsockname(plug, (struct sockaddr *)&socketname, &size);
  myaddr = ntohl(socketname.sin_addr.s_addr);

  #ifdef IRC_DEBUG
    cout << "Connect successful!" << endl;
  #endif

  Log::Write("Connect successful");

  if (Config::GetServerPassword() != "")
    Write("PASS " + Config::GetServerPassword() + "\n");
  Write("USER " + Config::GetUserName() + " oomon " + std::string(OOMON_VERSION) + " :" + Config::GetIRCName() + "\n");
  Write("NICK " + Config::GetNick() + "\n");
  MyNick = Config::GetNick();
  AmIAnOper = false;

  return plug;
}

int IRC::Disconnect(const std::string & Message)
{
  if (connected) {
    Write("QUIT :" + Message + "\n");
    #ifdef IRC_DEBUG
      cout << "Disconnecting from server." << endl;
    #endif
    Log::Write("Disconnecting from server");
    close(plug);
  }
  connected = plug = AmIAnOper = 0;
}

int IRC::Write(const std::string & msg)
{
  if (connected) {
    #ifdef IRC_DEBUG
      cout << "IRC << " << msg;
    #endif
    if (send(plug, msg.c_str(), msg.length(), 0) >= 0)
      return 0;
    else
      connected = false;
  }
  return INVALID;
}

void IRC::SetFD(fd_set & theset)
{
  if (connected)
    FD_SET(plug, &theset);
}

int IRC::Process(const fd_set & theset)
{
  std::string text;

  if (FD_ISSET(plug, &theset)) {
    if (int i = ReadLine(plug, text) <= 0)
      return INVALID;
    if (text != "") {
      #ifdef IRC_DEBUG
        cout << "IRC >> " << text << endl;
      #endif
      Parse(text);
    }
  }
  return 0;
}

void IRC::Parse(std::string text)
{
  StrVector Params;

  if (text == "")
    return;

  if (text[0] != ':') {
    SplitIRC(Params, text);
    Params.insert(Params.begin(), Server_Name);
  } else {
    SplitIRC(Params, text.substr(1));
  }
  int Numeric;
  std::string From, UserHost;
  std::string Command;

  SplitFrom(Params[0], From, UserHost);
  Command = Params[1];

  if ((Numeric = atoi(Command.c_str())) > 0)
  {
    switch(Numeric) {
    case 001:
      Server_Name = From;
      // Oper up and get O: lines
      if (!AmIAnOper)
      {
	Write("OPER " + Config::GetOperNick() + " " + Config::GetOperPass() +
	  "\n");
      }
      Klines.Clear();
      Dlines.Clear();
      GettingDlines = true;
      Write("STATS D\n");
      GettingKlines = true;
      Write("STATS K\n");
      if (Config::GetChannels() != "")
      {
        Write("JOIN " + Config::GetChannels() + "\n");
      }
      InitHash();
      Init_Nick_Change_Table();
      Init_Link_Look_Table();
      break;
    case 204:	// ontraceuser(body)
    case 205:	// ontraceuser(body)
      if (Params.size() > 6)
      {
        if ((Params[6][0] == '[') && (Params.size() > 7))
	{
          // [hybrid7-rc6] IRC >> :plasma.toast.pc 204 OOMon Oper opers OOMon [toast@Plasma.Toast.PC] (192.168.1.1) 0 0
          std::string Type = Params[3];
          std::string Class = Params[4];
          std::string Nick = Params[5];
          std::string UserHost = Params[6];
	  std::string IP = Params[7];
          OnTraceUser(Type, Class, Nick, UserHost, IP);
	}
	else
	{
          // [hybrid6] IRC >> :plasma.toast.pc 205 OOMon User 1 Toast[toast@Plasma.Toast.PC] (192.168.1.1) 000000005 000000005
          std::string Type = Params[3];
          std::string Class = Params[4];
          std::string NUH = Params[5];
	  std::string IP = Params[6];
          OnTraceUser(Type, Class, NUH, IP);
	}
      }
      break;
    case 209:	// ontraceclass()
      OnTraceClass();
      break;
    case 216:
// IRC >> :plasma.engr.arizona.edu 216 OOMon K *.monkeys.org * * foo llalalala (1998/03/03 11.18)
// IRC >> :plasma.engr.arizona.edu 216 OOMon K *bork.com * *hork moo la la la (1998/03/03 11.18)
      if (Params[3] == "K")
      {
	std::string host = Params[4];
	// Param[5] == (unused)
	std::string user = Params[6];
	std::string reason = (Params.size() > 7) ? Params[7] : "";
	for (StrVector::size_type pos = 8; pos < Params.size(); pos++)
	{
	  reason += " " + Params[pos];
	}
	Klines.Add(user + "@" + host, reason);
      }
      break;
    case 219:
      if (Params[3] == "K")
      {
        GettingKlines = false;
      }
      else if (Params[3] == "D")
      {
        GettingDlines = false;
      }
      break;
    case 225:
// IRC >> :plasma.toast.pc 225 ToastFOO D 1.2.3.4 :foo (2003/1/9 17.27)
      if (Params[3] == "D")
      {
	std::string host = Params[4];
	std::string reason = (Params.size() > 5) ? Params[5] : "";
	for (StrVector::size_type pos = 6; pos < Params.size(); pos++)
	{
	  reason += " " + Params[pos];
	}
	Dlines.Add(host, reason);
      }
      break;
#ifdef CA_SERVICES
    case 303:
      {
        std::string stuff = Params[3];
        if (stuff == "CA-SVCS ") {
	  if (!GettingCASVCSReports) {
	    Msg("CA-SVCS", "ADDNOTIFY HERE");
	    GettingCASVCSReports = true;
	  }
        } else {
	  GettingCASVCSReports = false;
        }
      }
      break;
#endif
    case 381:
      {
	AmIAnOper = true;
	#ifdef USE_FLAGS
	  Write("FLAGS +ALL\n");
	#endif
	Write("MODE " + MyNick + " :+szwckyrn\n");
      }
    }
  }
  else
  {
    switch(IRCCommand(Command)) {
    case IRC_PING:
      Write("PONG " + Params[2] + "\n");
      break;
    case IRC_NOTICE:
      {
	std::string To = Params[2];
	std::string Text = Params[3];

	OnNotice(From, UserHost, To, Text);
      }
      break;
    case IRC_PRIVMSG:
      {
	std::string To = Params[2];
	std::string Text = Params[3];

        OnPrivmsg(From, UserHost, To, Text);
      }
      break;
    case IRC_WALLOPS:
      {
	std::string text = Params[2];
	std::string wallopsType = "WALLOPS";

	if ((text.length() > 11) && (text.substr(0, 11) == "OPERWALL - "))
	{
	  wallopsType = "OPERWALL";
	  text = text.substr(11, std::string::npos);
	}
	else if ((text.length() > 9) && (text.substr(0, 9) == "LOCOPS - "))
	{
	  wallopsType = "LOCOPS";
	  text = text.substr(9, std::string::npos);
	}
	else if ((text.length() > 10) && (text.substr(0, 10) == "WALLOPS - "))
	{
	  text = text.substr(10, std::string::npos);
	}

        DCC::WriteByFlags(std::string("[") + From + ":" + wallopsType + "] " +
	  text + "\n", UF_WALLOPS);
      }
      break;
    }
  }
}

void IRC::OnCTCP(const std::string & From, const std::string & UserHost,
  const std::string & To, std::string Text)
{
  std::string command = FirstWord(Text);

#ifdef ONLY_OPERS_MAY_DCC
  if (Config::IsOper(UserHost) && (command == std::string("DCC")))
#else
  if (command == std::string("DCC"))
#endif
  {
    std::string dccCommand = FirstWord(Text);
    if (dccCommand == "CHAT")
    {
      FirstWord(Text);	// Ignore "chat" parameter

      unsigned long address = atoul(FirstWord(Text).c_str());;
      int port = atoi(FirstWord(Text).c_str());

      if (address == 0)
      {
        Notice(From, "Invalid address specified for DCC CHAT. Not funny.");
        return;
      }
      if (port < 1024)
      {
        Notice(From, "Invalid port specified for DCC CHAT. Not funny.");
        return;
      }
      Log::Write("DCC CHAT request from " + From + " (" + UserHost + ")");
      if (DCC::Connect(address, port, From, UserHost) != 0) {
        Notice(From, "DCC CHAT connection failed.");
        Log::Write("DCC CHAT failed for " + From);
      }
    }
  }
  else if (Config::IsOper(UserHost) && (command == std::string("CHAT")))
  {
    Log::Write("CHAT request from " + From + " (" + UserHost + ")");
    if (DCC::Listen(From, UserHost) != 0)
    {
      Notice(From, "DCC CHAT listen failed.");
      Log::Write("DCC CHAT listen failed for " + From);
    }
  }
  else if ((command == std::string("PING")) ||
    (command == std::string("ECHO")))
  {
    CTCPreply(From, command + " " + Text);
  }
  else if (command == std::string("VERSION"))
  {
    CTCPreply(From, std::string("VERSION OOMon") + OOMON_VERSION);
  }
}

void IRC::OnPrivmsg(const std::string & From, const std::string & UserHost,
  const std::string & To, std::string Text)
{
  if (Text[0] == '\001')
  {
    std::string ctcp = Text.substr(1);

    std::string::size_type end = ctcp.find('\001');

    if (end != std::string::npos)
    {
      OnCTCP(From, UserHost, To, ctcp.substr(0, end));
    }
  }
}

void IRC::OnNotice(const std::string & From, const std::string & UserHost,
	const std::string & To, std::string Text)
{
  if (Same(From, Server_Name)) {
    if (Text.length() > 14)
      if (Text.substr(0, 14) == "*** Notice -- ") {
	// Server Notice
        Text = Text.substr(14);

	if (Text.find("Client connecting: ") == 0) {
	  OnClientConnect(Text.substr(19));
	} else if (Text.find("Client exiting: ") == 0) {
	  OnClientExit(Text.substr(16));
	} else if (Text.find("Received KILL message for ") == 0) {
          Kill_Add_Report(Text.substr(26));
	} else if (Text.find("Rejecting clonebot:") == 0) {
	  CS_Clones(Text.substr(20));
	} else if (Text.find("Nick change:") == 0) {
	  Check_Nick_Flood(Text.substr(12));
	} else if (Text.find("Nick flooding detected by:") == 0) {
	  CS_Nick_Flood(Text.substr(26));
	} else if (Text.find("Rejecting ") == 0) {
	  Bot_Reject(Text.substr(10));
	} else if (Text.find("Clonebot killed:") == 0) {
	  CS_Clones(Text.substr(16));
	} else if (UpCase(Text).find("LINKS ") == 0) {
	  Link_Look_Notice(Text.substr(6));
#ifdef TRACE_BASTARD
	} else if (UpCase(Text).find("TRACE ") == 0) {
	  Trace_Notice(Text.substr(6));
#endif
	} else if (UpCase(Text).find("MOTD ") == 0) {
	  MOTD_Notice(Text.substr(5));
	} else if (UpCase(Text).find("INFO ") == 0) {
	  Info_Notice(Text.substr(5));
	} else if (UpCase(Text).find("STATS ") == 0) {
	  Stats_Notice(Text.substr(6));
	}
	else if (std::string::npos != Text.find("added K-Line for"))
        {
	  // Monitor and report added k-lines
	  Klines.ParseAndAdd(Text);
	}
	else if (std::string::npos != Text.find("added D-Line for"))
        {
	  // Monitor and report added d-lines
	  Dlines.ParseAndAdd(Text);
	}
	else if (std::string::npos !=
	  Text.find(" has removed the K-Line for: "))
	{
	  // Monitor and report removed k-lines
	  Klines.ParseAndRemove(Text);
	}
	else if (std::string::npos !=
	  Text.find(" has removed the D-Line for: "))
	{
	  // Monitor and report removed d-lines
	  Dlines.ParseAndRemove(Text);
	}
	else if ((std::string::npos != Text.find("K-Line for")) ||
	  (std::string::npos != Text.find("D-Line for")))
	{
	  // Not sure what this is for anymore!
	  ::Status(Text, SEND_OPERS_ONLY);
	}
	else if ((std::string::npos != Text.find("KLINE active for")) ||
	  (std::string::npos != Text.find("DLINE active for")))
	{
          // IRC >> :plasma.toast.pc NOTICE OOMon :*** Notice -- KLINE active for ToastRR[toast@cs6669210-179.austin.rr.com]
	  // Show activated k-lines and d-lines (disabled)
	  //DCC::Status(Text, SEND_OPERS_ONLY);
	  Log::Write(Text);
	}
      } else
      // plasma.engr.arizona.edu *** You need oper and N flag for +n
      if (UpCase(Text) == "*** YOU NEED OPER AND N FLAG FOR +N") {
	Log::Write("I don't have an N flag in my o: line! :(");
      }
#ifdef SERVICES
  } else if (Same(From, SERVICES_NAME)) {
    On_Services_Notice(Text);
#endif
#ifdef CA_SERVICES
  } else if (Same(From, CA_SERVICES_NAME)) {
    On_CA_Services_Notice(Text);
#endif
  }
}

void IRC::Notice(const std::string & To, const std::string & Text)
{
  Write("NOTICE " + To + " :" + Text + "\n");
}

void IRC::Msg(const std::string & To, const std::string & Text)
{
  Write("PRIVMSG " + To + " :" + Text + "\n");
}

void IRC::CTCP(const std::string & To, const std::string & Text)
{
  Write("PRIVMSG " + To + " :\001" + Text + "\001\n");
}

void IRC::CTCPreply(const std::string & To, const std::string & Text)
{
  Write("NOTICE " + To + " :\001" + Text + "\001\n");
}

void IRC::IsOn(const std::string & Text)
{
  Write("ISON " + Text + "\n");
}

void IRC::Kline(const std::string & From, const int Minutes,
	const std::string & Target, const std::string & Reason)
{
  std::string line;

  if (Minutes)
  {
    char buff[MAX_BUFF];
    sprintf(buff, "%u", Minutes);
    line = "KLINE " + std::string(buff) + " " + Target + " :" + Reason;
  }
  else
  {
    line = "KLINE " + Target + " :" + Reason;
  }

#ifdef OPER_NICK_IN_REASON
  Write(line + " [" + From + "]\n");
#else
  Write(line + "\n");
#endif
  Log::Write(line + " [" + From + "]");
}

void IRC::Unkline(const std::string & From, const std::string & Target)
{
  Write("UNKLINE " + Target + "\n");
  Log::Write("UNKLINE " + Target + " [" + From + "]");
}

void IRC::Dline(const std::string & From, const std::string & Target,
	const std::string & Reason)
{
  std::string line = "DLINE " + Target + " :" + Reason;
#ifdef OPER_NICK_IN_REASON
  Write(line + " [" + From + "]\n");
#else
  Write(line + "\n");
#endif
  Log::Write(line + " [" + From + "]");
}

void IRC::Undline(const std::string & From, const std::string & Target)
{
  Write("UNDLINE " + Target + "\n");
  Log::Write("UNDLINE " + Target + " [" + From + "]");
}

void IRC::Kill(const std::string & From, const std::string & Target,
  const std::string & Reason)
{
  std::string line = "KILL " + Target;
#ifdef OPER_NICK_IN_REASON
  Write(line + " :(" + From + ") " + Reason + "\n");
#else
  Write(line + " :" + Reason + "\n");
#endif
  Log::Write(line + " ((" + From + ") " + Reason + ")");
}

void IRC::Trace(const std::string & Target)
{
  if (Target == "")
    Write("TRACE\n");
  else
    Write("TRACE " + Target + '\n');
}

void IRC::StatsL(const std::string & Nick)
{
  if (Nick != "")
    Write("STATS L " + Nick + '\n');
}

void IRC::Retrace(const std::string & From)
{
  ::Status("Reload USERS requested by " + From, SEND_OPERS_ONLY);
  FreeHash();
  InitHash();
}

void IRC::ReloadKlines(const std::string & From)
{
  ::Status("Reload KLINES requested by " + From, SEND_OPERS_ONLY);
  Klines.Clear();
  GettingKlines = true;
  Write("STATS K\n");
}

void IRC::ReloadDlines(const std::string & From)
{
  ::Status("Reload DLINES requested by " + From, SEND_OPERS_ONLY);
  Dlines.Clear();
  GettingDlines = true;
  Write("STATS D\n");
}

std::string IRC::GetServerName()
{
  return Server_Name;
}

void IRC::FindK(StrList & Output, const std::string & userhost)
{
  Klines.Find(Output, userhost);
}

void IRC::FindD(StrList & Output, const std::string & userhost)
{
  Dlines.Find(Output, userhost);
}

unsigned long IRC::GetHostAddr()
{
  return HostAddr;
}

void IRC::Op(const std::string & OpChannel, const std::string & OpNick)
{
  Write("MODE " + OpChannel + " +o " + OpNick + "\n");
}

void IRC::Join(const std::string & Channel, const std::string & Key)
{
  Write("JOIN " + Channel + " " + Key + "\n");
}

void IRC::Part(const std::string & Channel)
{
  Write("PART " + Channel + "\n");
}

