// $Id: proxy.cc,v 1.19 2003/01/16 20:50:02 toast Exp $

// C headers
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <time.h>
#include <netdb.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <pwd.h>

// C++ headers
#include <iostream>
#include <fstream>
#include <string>
#include "oomon.h"
#include "config.h"
#include "log.h"
#include "proxy.h"
#include "socks4.h"
#include "wingate.h"
#include "util.h"
#include "main.h"
#include "engine.h"
#include "irc.h"


#ifndef INADDR_NONE
# define INADDR_NONE 0xffffffff
#endif


int Proxy::count = 0;
class Proxy* Proxy::items[MAX_PROXIES];

// Connect(address)
//
// Makes a client connection to an IP's proxy port.
//
// Returns 0 if no errors ocurred
//
int Proxy::Connect(const std::string & address, ProxyType type,
  const std::string & hostname, const std::string & nick,
  const std::string & userhost)
{
  struct sockaddr_in	socketname;
  struct sockaddr_in	localaddr;
  struct hostent	*remote_host;
  unsigned long		addr;
  char			buf[MAX_BUFF];
  int			optval;
  int			plug;
  int			port = 0;

  if (type == WINGATE)
  {
    port = 23;
  }
  else if (type == SOCKS4)
  {
    port = 1080;
  }

  if (Detecting(address, type))
    return INVALID;

  if (count >= MAX_PROXIES) {
    Log::Write("Increase MAX_PROXIES!");
    return INVALID;
  }

#ifdef PROXY_DEBUG
    cout << "Creating proxy connection to address " << address << ":" <<
      port << endl;
#endif

  if ((plug = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    cerr << "error: can't assign fd for socket in Proxy::Connect" << endl;
    return INVALID;
  } 

  (void)setsockopt(plug, SOL_SOCKET, SO_REUSEADDR, (char *)&optval,
        sizeof(optval));

  (void)bzero((char *)&localaddr, sizeof(struct sockaddr_in));
  if (!(localaddr.sin_addr.s_addr = Config::GetWGvhost()))
    localaddr.sin_addr.s_addr = htonl(IRC::GetMyAddr());
  localaddr.sin_family = AF_INET;
  localaddr.sin_port = 0;

  if (bind(plug, (struct sockaddr *) &localaddr, sizeof(localaddr)) < 0)
    cerr << "Error: Bind " << errno << " (" << strerror(errno) << ")" << endl;

  optval = 1;
  (void)setsockopt(plug, SOL_SOCKET, SO_REUSEADDR, (char *)&optval,
                   sizeof(optval));
  (void)bzero((char *)&socketname, sizeof(socketname));

  if (INADDR_NONE == (addr = inet_addr(address.c_str()))) {
    cerr << "error: invalid address \"" << address << "\" passed to Proxy::Connect" << endl;
    return INVALID;
  }

  socketname.sin_addr.s_addr=addr;
  socketname.sin_family = AF_INET;
  socketname.sin_port = htons(port);

  if (-1 == fcntl(plug, F_SETFL, O_NONBLOCK)) {
    cerr << "error: can't set fd non-blocking in Proxy::Connect" << endl;
    return INVALID;
  }

  if (connect(plug, (struct sockaddr *) &socketname, sizeof(socketname)) < 0)
    if (errno != EINPROGRESS) {
#ifdef PROXY_DEBUG
        cout << "Error: Connect " << errno << " (" << strerror(errno) << ")" << endl;
#endif
      return INVALID;
    }
  
  if (type == WINGATE)
  {
    items[count] = new WinGate(plug, address, hostname, nick, userhost);
  }
  else if (type == SOCKS4)
  {
    items[count] = new Socks4(plug, address, hostname, nick, userhost);
  }
  count++;
#ifdef PROXY_DEBUG
    cout << count << " proxies queued up for processing." << endl;
#endif

  return 0;
}

void
Proxy::initialize(const int fd, const std::string & address,
  const std::string & hostname, const std::string & nick,
  const std::string & userhost)
{
  plug = fd;
  connected = false;
  gotdata = false;
  Address = address;
  Hostname = hostname;
  Nick = nick;
  Userhost = userhost;
  LastActive = time((time_t *)NULL);
}


// SetFD(readset, writeset)
//
// Do this if you want to be able to receive data :P
//
void Proxy::SetFD(fd_set & readset, fd_set & writeset)
{
  class Proxy *temp;

  for (int i = 0; i < count; i++) {
    temp = items[i];
    temp->PrivSetFD(readset, writeset);
  }
}


// PrivSetFD(readset, writeset)
//
// Duh!
//
void Proxy::PrivSetFD(fd_set & readset, fd_set & writeset) const
{
  if (connected) {
    FD_SET(plug, &readset);
  } else {
    FD_SET(plug, &writeset);
    FD_SET(plug, &readset);
  }
}


// Process(readset, writeset)
//
// Processes any received data at each proxy connection.
//
// Returns 0 if no errors ocurred.
//
int Proxy::Process(const fd_set & readset, const fd_set & writeset)
{
  class Proxy *temp;
  int i = 0;
  while (i < count) {
    temp = items[i];
    if (!temp->PrivProcess(readset, writeset))
    {
      delete temp;
#ifdef PROXY_DEBUG
	cout << "Lost proxy connection" << endl;
#endif
      for (int j = i; j < count-1; j++)
        items[j] = items[j+1];
      count--;
#ifdef PROXY_DEBUG
        cout << count << " proxies queued up for processing." << endl;
#endif
    } else
      i++;
  }
  return 0;
}


// PrivProcess(readset, writeset)
//
// Processes any incoming data or connections
//
// Returns true if no errors ocurred
//
bool Proxy::PrivProcess(const fd_set & readset, const fd_set & writeset)
{
  static char buff[2048];
  unsigned short port;
  unsigned long dst;
  int m, n;

  if (FD_ISSET(plug, &readset))
  {
    bzero((char *)&buff, sizeof(buff));
    if ((n = read(plug, &buff, sizeof(buff))) > 0)
    {
      gotdata = true;
      LastActive = time((time_t *)NULL);
#ifdef PROXY_DEBUG
        cout << "Proxy >> " << buff << endl;
#endif
      return this->Parse(buff);
    }
    else if (n == 0)
    {
      return false;
    }
    else
    {
#ifdef PROXY_DEBUG
        cout << "Error: Read " << errno << " (" << strerror(errno) << ")" << endl;
#endif
      return false;
    }
  }

  if (!connected)
  {
    if (FD_ISSET(plug, &writeset))
    {
      connected = true;
      LastActive = time((time_t *)NULL);
      if (!this->OnConnect())
      {
	return false;
      }
    }
  }

  if (GetIdle() > PROXY_IDLE_MAX) {
#ifdef PROXY_DEBUG
      cout << "Proxy connection timed out for " << GetAddress() << endl;
#endif
    return false;
  }
  return true;
}


// GetIdle()
//
// Returns the amount of time the connection has been idle
//
int Proxy::GetIdle() const
{
  time_t CurrentTime;

  CurrentTime = time((time_t *)NULL);
  return (CurrentTime - LastActive);
}

void Proxy::Disconnect()
{
  if (plug > 0)
    close(plug);
  plug = 0;
}

int Proxy::Write(char *buffer, int size)
{
  return (write(plug, buffer, size));
}

std::string Proxy::GetAddress() const
{
  return Address;
}

std::string Proxy::GetHostname() const
{
  return Hostname;
}

std::string Proxy::GetNick() const
{
  return Nick;
}

std::string Proxy::GetUserhost() const
{
  return Userhost;
}

bool Proxy::Detecting(const std::string & address, const ProxyType type)
{
  for (int i = 0; i < count; i++)
    if ((items[i]->GetAddress() == address) && (items[i]->GetType() == type))
      return true;
  return false;
}
