// $Id: trap.h,v 1.2 2003/01/10 23:02:28 toast Exp $

#ifndef _TRAP_H_
#define _TRAP_H_

#include <string>
#include <list>

#include "strtype"


enum TrapAction
{
  TRAP_ECHO, TRAP_KILL, TRAP_KLINE, TRAP_KLINEHOST, TRAP_DLINE, TRAP_DLINENET
};


class Trap
{
private:
  TrapAction	action;
  long		timeout;	// For K-Lines only
  std::string	nick;
  std::string	userhost;
  std::string	reason;		// For Kills, K-Lines, and D-Lines only

public:
  Trap(const TrapAction action, const long timeout, const std::string & nick,
    const std::string & userhost, const std::string & reason);

  bool matches(const std::string & nick, const std::string & userhost) const;
  bool exactMatch(const std::string & nick, const std::string & userhost) const;

  TrapAction getAction() const;
  long getTimeout() const;
  std::string getMask() const;
  std::string getReason() const;
};


class TrapList
{
private:
  static std::list<Trap> traps;

  static void add(const TrapAction action, const long timeout,
    const std::string & mask, const std::string & reason);

  static void split(const std::string & mask, std::string & nick,
    std::string & userhost);

public:
  static void addEcho(const std::string & mask);
  static void addKill(const std::string & mask, const std::string & reason);
  static void addKline(const bool host, const std::string & mask,
    const std::string & reason, const long timeout = 0);
  static void addDline(const bool net, const std::string & mask,
    const std::string & reason);
  static bool remove(const std::string & mask);
  static bool remove(const std::string & nick, const std::string & userhost);
  static void flush(void);

  static void match(const std::string & nick, const std::string & userhost,
    const std::string & ip);

  static void list(StrList & output);
};

#endif

