// $Id: wingate.cc,v 1.20 2003/01/16 21:35:10 toast Exp $

// Std C++ headers
#include <iostream>
#include <fstream>
#include <string>

// OOMon headers
#include "oomon.h"
#include "wingate.h"
#include "log.h"
#include "main.h"
#include "engine.h"
#include "irc.h"


WinGate::WinGate(const int fd, const std::string & address,
  const std::string & hostname, const std::string & nick,
  const std::string & userhost)
{
  this->initialize(fd, address, hostname, nick, userhost);
}


// OnConnect()
//
// Returns true if no errors occurred
//
bool WinGate::OnConnect()
{
#ifdef PROXY_DEBUG
  cout << "WinGate detector connected to " << GetAddress() << endl;
#endif

  return true;
}


// Parse(text)
//
// Parses incoming data
//
// Returns true if no errors occurred
//
bool WinGate::Parse(std::string text)
{
  if ((std::string::npos != text.find("WinGate>")) || (std::string::npos !=
    text.find("Too many connected users - try again later")))
  {
#ifdef PROXY_DEBUG
    cout << "WINGATE DETECTED!" << endl;
#endif
    std::string notice = std::string("WinGate detected for ") + GetNick() +
      "!" + GetUserhost() + " [" + GetAddress() + "]";
    ::Status(notice, SEND_OPERS_ONLY);
    Log::Write(notice);

#if defined(AUTO_KLINE_WINGATES) || defined(AUTO_DLINE_WINGATES)
    if (GetAutoPilot())
    {
# ifdef AUTO_DLINE_WINGATES
      IRC::Dline("Auto-Dline", GetAddress(), REASON_WINGATE);
# else
#  ifdef KLINE_IP_SUPPORTED
      IRC::Kline("Auto-Kline", WINGATE_KLINE_TIME, "*@" + GetAddress(),
        REASON_WINGATE);
#  else
      IRC::Kline("Auto-Kline", WINGATE_KLINE_TIME, "*@" + GetHostname(),
        REASON_WINGATE);
#  endif
# endif
    }
    else
#endif
    {
      ::Status(::Make_Kline("*@" + GetHostname(), MK_WINGATE,
        WINGATE_KLINE_TIME), SEND_KLINERS_ONLY);
    }
    return false;
  }
  return true; 
}

